# -*- coding: utf-8 -*-

""" Checks all data resources give 200s
"""

from unittest import main, TestCase
import re
import os
import json

import requests

from g2p.app import APP
from g2p.log import LOGGER
from g2p.mappings.langs import LANGS_NETWORK
from g2p.tests.public import __file__ as PUB_FILE

PUB_DIR = os.path.dirname(PUB_FILE)

class ResourceIntegrationTest(TestCase):
    """
    This tests that the api returns 200s for all basic
    GET requests.
    """

    def setUp(self):
        # Test external hosts
        self.client = APP.test_client
        self.prefix = "/api/v1"
        # routes
        self.conversion_route = '/api/v1/g2p'
        self.static_route = "/static/<path:filename>"
        self.routes = [str(route) for route in APP.url_map.iter_rules()]
        self.routes_no_args = [
            route for route in self.routes if not "<" in route and route != self.conversion_route]
        self.routes_only_args = [
            route for route in self.routes if "<" in route and route != self.static_route]
        # endpoints
        self.rules_by_endpoint = APP.url_map._rules_by_endpoint
        self.endpoints = [rt for rt in self.rules_by_endpoint.keys()]
        # args
        self.arg_match = re.compile(r'\<[a-z:]+\>')
        self.args_to_check = ("node")

    def return_endpoint_arg(self, ep):
        split = ep.split('.')
        split_length = len(split)
        return split[split_length-1]

    def return_route_from_endpoint(self, ep):
        return str(self.rules_by_endpoint[ep][0])

    def test_response_code(self):
        '''
        Ensure all routes return 200
        '''
        for rt in self.routes_no_args:
            try:
                r = self.client().get(rt)
                self.assertEqual(r.status_code, 200)
                LOGGER.debug("Route " + rt +
                             " returned " + str(r.status_code))
            except:
                LOGGER.error("Couldn't connect. Is flask running?")

    def test_response_code_with_args(self):
        '''
        Ensure all args return 200
        '''
        for ep in self.routes_only_args:
            for node in LANGS_NETWORK.nodes:
                rt = re.sub(self.arg_match, node, ep)
                try:
                    r = self.client().get(rt)
                    self.assertEqual(r.status_code, 200)
                except:
                    LOGGER.error("Couldn't connect. Is flask running?")
            LOGGER.debug("Successfully tested " + str(len(LANGS_NETWORK.nodes)
                                                      ) + " node resources at route " + ep + " .")

    def test_g2p_conversion(self):
        '''
        Ensure conversion returns proper response
        '''
        params = {'in-lang': 'dan', 'out-lang': 'eng-arpabet',
                  'text': "hej", 'debugger': True, 'index': True}
        bad_params = {'in-lang': 'dan', 'out-lang': 'moh', 'text': "hej"}
        missing_params = {'in-lang': 'not-here',
                          'out-lang': 'eng-arpabet', 'text': "hej"}
        self.maxDiff = None
        response = self.client().get(self.conversion_route, query_string=params)
        res_json = response.get_json()
        self.assertEqual(response.status_code, 200)
        with open(os.path.join(PUB_DIR, 'sample_response.json')) as f:
            data = json.load(f)
        self.assertEqual(
            res_json, data)
        bad_response = self.client().get(self.conversion_route, query_string=bad_params)
        self.assertEqual(bad_response.status_code, 400)
        missing_response = self.client().get(
            self.conversion_route, query_string=missing_params)
        self.assertEqual(missing_response.status_code, 404)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

######################################################################
# © Patrick Littell
#
# create_ipa_mapping.py
#
# Given two IPA inventories in JSON (either as dedicated inventory
# files or the input/output sides of mapping files), map the first
# onto the second by use of panphon's phonetic distance calculators.
#
# The resulting mappings are used just like other mappings: to make
# converters and pipelines of converters in convert_orthography.py
#
# AP Note: Taken from ReadAlongs-Studio and implemented with G2P formatting
######################################################################

from __future__ import print_function, unicode_literals
from __future__ import division, absolute_import

from copy import deepcopy
import json
import os

from panphon.xsampa import XSampa
import panphon.distance
from tqdm import tqdm
import yaml

from g2p.mappings.utils import is_ipa, is_xsampa, IndentDumper
from g2p.mappings import Mapping
from g2p.log import LOGGER

#################################
#
# Preprocessing:
#
# Panphon can only match a single segment to another segment, rather
# than (say) try to combine two segments to better match the features.
# For example, you might want "kʷ" to match to "kw", but Panphon will
# only match the "kʷ" to "k" and consider the "w" to be a dropped
# character.  In order to get around this, we preprocess strings so
# that common IPA segments that you might expect map to two characters
# in another language, like affricates or rounded consonants, are
# treated as two characters rather than one.
#
#################################
xsampa_converter = XSampa()


def process_character(p, is_xsampa=False):
    if is_xsampa:
        p = xsampa_converter.convert(p)
    return p.replace("ʷ", "w").replace("ʲ", "j").replace("͡", "")


def process_characters(inv, is_xsampa=False):
    return [process_character(p, is_xsampa) for p in inv]

##################################
#
# Creating the mapping
#
#
#
###################################


def create_mapping(mapping_1: Mapping, mapping_2: Mapping, mapping_1_io: str = 'out', mapping_2_io: str = 'in', write_to_file: bool = False, out_dir: str = '') -> Mapping:
    map_1_name = mapping_1.kwargs[f'{mapping_1_io}_lang']
    map_2_name = mapping_2.kwargs[f'{mapping_2_io}_lang']
    if not is_ipa(map_1_name) and not is_xsampa(map_1_name):
        LOGGER.warning("Unsupported orthography of inventory 1: %s"
                       " (must be ipa or x-sampa)",
                       map_1_name)
    if not is_ipa(map_2_name) and not is_xsampa(map_2_name):
        LOGGER.warning("Unsupported orthography of inventory 2: %s"
                       " (must be ipa or x-sampa)",
                       map_2_name)
    l1_is_xsampa, l2_is_xsampa = is_xsampa(map_1_name), is_xsampa(map_2_name)
    mapping = align_inventories(mapping_1.inventory(mapping_1_io), mapping_2.inventory(mapping_2_io),
                                l1_is_xsampa, l2_is_xsampa)

    # Initialize mapping with input language parameters (as_is,
    # case_sensitive, prevent_feeding, etc)
    config = mapping_1.kwargs.copy()
    # Fix up names, etc.
    if 'authors' in config:
        del config['authors']
    if 'display_name' in config:
        del config['display_name']
    if 'language_name' in config:
        del config['language_name']
    config['in_lang'] = map_1_name
    config['out_lang'] = map_2_name
    config['mapping'] = mapping
    mapping = Mapping(**config)
    if write_to_file:
        if out_dir:
            if os.path.isdir(out_dir):
                mapping.config_to_file(out_dir)
                mapping.mapping_to_file(out_dir)
            else:
                LOGGER.warning(f'{out_dir} is not a directory. Writing to default instead.')
        else:
            mapping.config_to_file()
            mapping.mapping_to_file()

    return mapping

def find_good_match(p1, inventory_l2, l2_is_xsampa=False):
    """Find a good sequence in inventory_l2 matching p1."""

    dst = panphon.distance.Distance()
    # The proper way to do this would be with some kind of beam search
    # through a determinized/minimized FST, but in the absence of that
    # we can do a kind of heurstic greedy search.  (we don't want any
    # dependencies outside of PyPI otherwise we'd just use OpenFST)
    p1_pseq = dst.fm.ipa_segs(p1)
    p2_pseqs = [dst.fm.ipa_segs(p)
                for p in process_characters(inventory_l2, l2_is_xsampa)]
    i = 0
    good_match = []
    while i < len(p1_pseq):
        best_input = ""
        best_output = -1
        best_score = 0xdeadbeef
        for j, p2_pseq in enumerate(p2_pseqs):
            # FIXME: Should also consider the (weighted) possibility
            # of deleting input or inserting any segment (but that
            # can't be done with a greedy search)
            if len(p2_pseq) == 0:
                LOGGER.warning('No panphon mapping for %s - skipping',
                               inventory_l2[j])
                continue
            e = min(i + len(p2_pseq), len(p1_pseq))
            input_seg = p1_pseq[i:e]
            score = dst.weighted_feature_edit_distance(''.join(input_seg),
                                                       ''.join(p2_pseq))
            # Be very greedy and take the longest match
            if (score < best_score
                or score == best_score
                    and len(input_seg) > len(best_input)):
                best_input = input_seg
                best_output = j
                best_score = score
        LOGGER.debug('Best match at position %d: %s => %s',
                     i, best_input, inventory_l2[best_output])
        good_match.append(inventory_l2[best_output])
        i += len(best_input)  # greedy!
    return ''.join(good_match)


def align_inventories(inventory_l1, inventory_l2,
                      l1_is_xsampa=False, l2_is_xsampa=False):
    mapping = []
    pbar = tqdm(total=100)
    step = 1/len(inventory_l1)*100
    for i1, p1 in enumerate(process_characters(inventory_l1,
                                               l1_is_xsampa)):
        # we enumerate the strings because we want to save the original string
        # (e.g., 'kʷ') to the mapping, not the processed one (e.g. 'kw')
        good_match = find_good_match(p1, inventory_l2, l2_is_xsampa)
        mapping.append({"in": inventory_l1[i1], "out": good_match})
        pbar.update(step)
    pbar.close()
    return mapping

if __name__ == '__main__':
    test_1 = Mapping(in_lang='atj', out_lang='atj-ipa')
    test_2 = Mapping(in_lang='eng-ipa', out_lang='eng-arpabet')
    create_mapping(test_1, test_2, write_to_file=True)



''' Very Basic API
'''
import os
import json

from flask import Blueprint, abort
from flask_restful import (Resource, Api, reqparse)
from flask_cors import CORS

from networkx.exception import NetworkXError, NetworkXNoPath
from networkx.algorithms.dag import ancestors, descendants
from g2p.static import __file__ as static_file
from g2p.transducer import Transducer
from g2p.mappings.langs import LANGS_NETWORK
from g2p.log import LOGGER
from g2p import make_g2p


class Ancestors(Resource):
    def get(self, node):
        try:
            return sorted([x for x in ancestors(LANGS_NETWORK, node)])
        except NetworkXError:
            abort(404)


class Descendants(Resource):
    def get(self, node):
        try:
            return sorted([x for x in descendants(LANGS_NETWORK, node)])
        except NetworkXError:
            abort(404)


class Langs(Resource):
    def get(self):
        return sorted([x for x in LANGS_NETWORK.nodes])


class Text(Resource):
    def __init__(self):
        self.parser = reqparse.RequestParser()
        self.parser.add_argument(
            'in-lang', dest='in-lang',
            type=str, location='args',
            required=True, help='The input language',
        )
        self.parser.add_argument(
            'out-lang', dest='out-lang',
            type=str, location='args',
            required=True, help="The output language",
        )
        self.parser.add_argument(
            'text', dest='text',
            type=str, location='args',
            required=True, help="The text in the input language",
        )
        self.parser.add_argument(
            'index', dest='index',
            type=bool, location='args',
            default=False, required=False,
            help="Return indices",
        )
        self.parser.add_argument(
            'debugger', dest='debugger',
            type=bool, location='args',
            default=False, required=False,
            help="The text in the input language",
        )

    def get(self):
        args = self.parser.parse_args()
        in_lang = args['in-lang']
        out_lang = args['out-lang']
        text = args['text']
        index = args['index']
        debugger = args['debugger']
        try:
            transducer = make_g2p(in_lang, out_lang)
            tg = transducer(text)
            text = tg.output_string
            input_text = tg.input_string
            if debugger:
                debugger = tg.debugger
            if index:
                index = tg.edges
            return {'input-text': input_text, 'output-text': text, 'index': index, 'debugger': debugger}
        except NetworkXNoPath:
            abort(400)
        except FileNotFoundError:
            abort(404)

def update_docs():
    ''' Update the swagger documentation with all nodes from the network
    '''
    swagger_path = os.path.join(os.path.dirname(static_file), 'swagger.json')
    with open(swagger_path) as f:
        data = json.load(f)
    data['components']['schemas']['Langs']['enum'] = sorted(
        [x for x in LANGS_NETWORK.nodes])
    with open(swagger_path, 'w') as f:
        f.write(json.dumps(data))
    LOGGER.info('Updated API documentation')


g2p_api = Blueprint('resources.g2p', __name__)

CORS(g2p_api)

api = Api(g2p_api)

api.add_resource(
    Ancestors,
    '/ancestors/<string:node>',
    endpoint='ancestors'
)

api.add_resource(
    Descendants,
    '/descendants/<string:node>',
    endpoint='descendants'
)

api.add_resource(
    Langs,
    '/langs',
    endpoint='langs'
)

api.add_resource(
    Text,
    '/g2p',
    endpoint='g2p'
)



"""

Basic init file for g2p module

"""
import sys
import io

from networkx import shortest_path
from networkx.exception import NetworkXNoPath

from g2p.mappings import Mapping
from g2p.mappings.langs import LANGS_NETWORK
from g2p.transducer import CompositeTransducer, Transducer
from g2p.log import LOGGER

if sys.stdout.encoding != 'utf8' and hasattr(sys.stdout, 'buffer'):
    sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding="utf8")

if sys.stderr.encoding != 'utf8' and hasattr(sys.stderr, 'buffer'):
    sys.stderr = io.TextIOWrapper(sys.stderr.buffer, encoding="utf8")

def make_g2p(in_lang: str, out_lang: str):
    # Check in_lang is a node in network
    if in_lang not in LANGS_NETWORK.nodes:
        LOGGER.error(f"No lang called {in_lang}. Please try again.")
        raise(FileNotFoundError("No lang called {in_lang}."))

    # Check out_lang is a node in network
    if out_lang not in LANGS_NETWORK.nodes:
        LOGGER.error(f"No lang called {out_lang}. Please try again.")
        raise(FileNotFoundError("No lang called {out_lang}."))

    # Try to find the shortest path between the nodes
    try:
        path = shortest_path(LANGS_NETWORK, in_lang, out_lang)
    except NetworkXNoPath:
        LOGGER.error(f"Sorry, we couldn't find a way to convert {in_lang} to {out_lang}. Please update your langs by running `g2p update` and try again.")
        raise(NetworkXNoPath)

    # Find all mappings needed
    mappings_needed = []
    for i, lang in enumerate(path):
        try:
            mapping = Mapping(in_lang=path[i], out_lang=path[i+1])
            LOGGER.debug(f"Adding mapping between {path[i]} and {path[i+1]} to composite transducer.")
            mappings_needed.append(mapping)
        except IndexError:
            continue

    # Either return Transducer or Composite Transducer
    if len(mappings_needed) == 1:
        return Transducer(mappings_needed[0])
    else:
        return CompositeTransducer([Transducer(x) for x in mappings_needed])

    



""" Test Mapping utility functions
"""

import os
from unittest import main, TestCase
from collections import defaultdict

import yaml

from g2p.mappings import utils
from g2p.mappings import Mapping
from g2p.tests.public import PUBLIC_DIR
from g2p.exceptions import IncorrectFileType, MalformedMapping


class UtilsTest(TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        gen_mapping = os.path.join(PUBLIC_DIR, 'mappings', 'test_to_test-out.json')
        gen_config = os.path.join(PUBLIC_DIR, 'mappings', 'test_config.yaml')
        if os.path.exists(gen_config):
            os.remove(gen_config)
        if os.path.exists(gen_mapping):
            os.remove(gen_mapping)
        fresh_config = {'language_name': 'generated', 'mappings': []}
        with open(os.path.join(PUBLIC_DIR, 'mappings', 'generated_add.yaml'), 'w') as f:
            yaml.dump(fresh_config, f, Dumper=utils.IndentDumper, default_flow_style=False)

    def test_abb_flatten_and_expand(self):
        test_rows = [
            ["VOWEL", 'a', 'e', 'i', 'o', 'u'],
            ["OTHER", 't', 'e', 's', 't']
        ]
        default_dict = defaultdict(list)
        default_dict['VOWEL'].extend(['a', 'e', 'i', 'o', 'u'])
        default_dict['OTHER'].extend(['t', 'e', 's', 't'])
        empty_rows = []
        while len(empty_rows) < 10:
            empty_rows.append(['', '', '', '', '', ''])
        self.assertEqual(utils.flatten_abbreviations(test_rows), default_dict)
        self.assertEqual(utils.expand_abbreviations(default_dict), test_rows)
        self.assertEqual(utils.expand_abbreviations({}), empty_rows)

    def test_unicode_escape(self):
        ''' Should turn \u0331 declared in CSVs
            into actual Unicode string for that codepoint
        '''
        self.assertEqual('\u0000', utils.unicode_escape('\\u0000'))
        self.assertEqual('\u0331', utils.unicode_escape('\\u0331'))
        self.assertEqual('\u26F0', utils.unicode_escape('\\u26F0'))

    def test_fixed_width(self):
        # TODO: Test utils.create_fixed_width_lookbehind and utils.pattern_to_fixed_width_lookbehinds
        pass

    def test_pattern(self):
        pass

    def test_load_mapping(self):
        with self.assertRaises(MalformedMapping):
            utils.load_mapping_from_path(os.path.join(
                PUBLIC_DIR, 'mappings', 'malformed_config.yaml'))
        minimal = utils.load_mapping_from_path(os.path.join(
            PUBLIC_DIR, 'mappings', 'minimal_config.yaml'))
        csv = utils.load_mapping_from_path(os.path.join(
            PUBLIC_DIR, 'mappings', 'minimal_configs.yaml'), 0)
        tsv = utils.load_mapping_from_path(os.path.join(
            PUBLIC_DIR, 'mappings', 'minimal_configs.yaml'), 1)
        psv = utils.load_mapping_from_path(os.path.join(
            PUBLIC_DIR, 'mappings', 'minimal_configs.yaml'), 2)
        json = utils.load_mapping_from_path(os.path.join(
            PUBLIC_DIR, 'mappings', 'minimal_configs.yaml'), 3)
        xlsx = utils.load_mapping_from_path(os.path.join(
            PUBLIC_DIR, 'mappings', 'minimal_configs.yaml'), 4)
        self.assertEqual(minimal['mapping_data'], csv['mapping_data'])
        self.assertEqual(minimal['mapping_data'], tsv['mapping_data'])
        self.assertEqual(minimal['mapping_data'], psv['mapping_data'])
        self.assertEqual(minimal['mapping_data'], json['mapping_data'])
        self.assertEqual(minimal['mapping_data'], xlsx['mapping_data'])

    def test_validate(self):
        pass

    def test_escape_special(self):
        self.assertEqual(utils.escape_special_characters({'in': '?'}), {'in': '\?'})

    def test_load_abbs(self):
        with self.assertRaises(IncorrectFileType):
            utils.load_abbreviations_from_file(os.path.join(
                PUBLIC_DIR, 'mappings', 'abbreviations.json'))
        for abb in ['abbreviations.csv', 'abbreviations.tsv', 'abbreviations.psv']:
            abbs = utils.load_abbreviations_from_file(
                os.path.join(PUBLIC_DIR, 'mappings', abb))
            self.assertTrue("VOWEL" in abbs)
            self.assertEqual(abbs['VOWEL'], ['a', 'e', 'i', 'o', 'u'])

    def test_generated_mapping(self):
        config = {'in_lang': 'test', 'out_lang': 'test-out', 'as_is': False}
        # config = utils.generate_config('test', 'test-out', 'Test', 'TestOut')
        config['mapping'] = [{'in': 'a', 'out': 'b'}]
        mapping = Mapping(**config)
        mapping.config_to_file(os.path.join(PUBLIC_DIR, 'mappings', 'test_config.yaml'))
        mapping.config_to_file(os.path.join(PUBLIC_DIR, 'mappings', 'generated_add.yaml'))
        mapping.mapping_to_file(os.path.join(PUBLIC_DIR, 'mappings'))
        test_config = utils.load_mapping_from_path(os.path.join(
            PUBLIC_DIR, 'mappings', 'test_config.yaml'))
        test_config_added = utils.load_mapping_from_path(os.path.join(
            PUBLIC_DIR, 'mappings', 'generated_add.yaml'))
        self.assertEqual(test_config['mapping_data'], [{'in': 'a', 'out': 'b', 'context_before': '', 'context_after': ''}])
        self.assertEqual(test_config['in_lang'], 'test')
        self.assertEqual(test_config['out_lang'], 'test-out')
        self.assertEqual(test_config['language_name'], 'test')
        self.assertEqual(test_config['display_name'], 'test custom to test-out custom')
        self.assertEqual(test_config_added['mapping_data'], [{'in': 'a', 'out': 'b', 'context_before': '', 'context_after': ''}])
        self.assertEqual(test_config_added['in_lang'], 'test')
        self.assertEqual(test_config_added['out_lang'], 'test-out')
        self.assertEqual(test_config_added['language_name'], 'test')
        self.assertEqual(test_config_added['display_name'], 'test custom to test-out custom')

if __name__ == '__main__':
    main()



"""

Views and config to the g2p Studio web app

"""
import json
import os
import requests
from networkx.algorithms.dag import ancestors, descendants
from networkx.drawing.layout import spring_layout, spectral_layout, shell_layout, circular_layout
from flask import Flask, render_template
from flask_cors import CORS
from flask_socketio import SocketIO, emit
from flask_talisman import Talisman
from typing import List, Union

from g2p.mappings import Mapping
from g2p.mappings.langs import LANGS, LANGS_NETWORK
from g2p.transducer import CompositeTransducer, Transducer, TransductionGraph, CompositeTransductionGraph
from g2p.static import __file__ as static_file
from g2p.mappings.utils import expand_abbreviations, flatten_abbreviations
from g2p.api import g2p_api
from g2p.log import LOGGER
from g2p import make_g2p

APP = Flask(__name__)
APP.register_blueprint(g2p_api, url_prefix='/api/v1')
CORS(APP)
SOCKETIO = SocketIO(APP)
DEFAULT_N = 10


def shade_colour(colour, percent, r=0, g=0, b=0):
    R = hex(
        min(255, int((int(colour[1:3], 16) * (100 + percent + r) / 100)))).lstrip('0x')
    G = hex(
        min(255, int((int(colour[3:5], 16) * (100 + percent + g) / 100)))).lstrip('0x')
    B = hex(
        min(255, int((int(colour[5:], 16) * (100 + percent + b) / 100)))).lstrip('0x')
    return '#' + str(R) + str(G) + str(B)


def contrasting_text_color(hex_str):
    (R, G, B) = (hex_str[1:3], hex_str[3:5], hex_str[5:])
    return '#000' if 1 - (int(R, 16) * 0.299 + int(G, 16) * 0.587 + int(B, 16) * 0.114) / 255 < 0.5 else '#fff'


def network_to_echart(write_to_file: bool = False, layout: bool = False):
    nodes = []
    no_nodes = len(LANGS_NETWORK.nodes)
    for node in LANGS_NETWORK.nodes:
        lang_name = node.split('-')[0]
        no_ancestors = len(ancestors(LANGS_NETWORK, node))
        no_descendants = len(descendants(LANGS_NETWORK, node))
        size = min(20, max(2, ((no_ancestors/no_nodes) *
                               100 + (no_descendants/no_nodes)*100)))
        node = {'name': node, 'symbolSize': size,
                'id': node, 'category': lang_name}
        nodes.append(node)
    edges = []
    for edge in LANGS_NETWORK.edges:
        edges.append({'source': edge[0], 'target': edge[1]})
    if write_to_file:
        with open(os.path.join(os.path.dirname(static_file), 'languages-network.json'), 'w') as f:
            f.write(json.dumps({'nodes': nodes, 'edges': edges}))
        LOGGER.info(f'Wrote network nodes and edges to static file.')
    return nodes, edges


def return_echart_data(tg: Union[CompositeTransductionGraph, TransductionGraph]):
    x = 100
    diff = 200
    nodes = []
    edges = []
    index_offset = 0
    colour = "#222222"
    if isinstance(tg, CompositeTransductionGraph):
        steps = len(tg.tiers)
    else:
        steps = 1
    for ind, tier in enumerate(tg.tiers):
        if ind == 0:
            symbol_size = min(300 / len(tier.input_string), 40)
            input_x = x + (ind * diff)
            input_y = 300
            x += diff
            inputs = [{'name': f"{x}",
                       'id': f"(in{ind+index_offset}-in{i+index_offset})",
                       'x': input_x,
                       'y': input_y + (i*50),
                       'symbolSize': symbol_size,
                       'label': {'color': contrasting_text_color(colour)},
                       'itemStyle': {'color': colour,
                                     'borderColor': contrasting_text_color(colour)}}
                      for i, x in enumerate(tier.input_string)]
            nodes += inputs
        edges += [{"source": x[0] + index_offset, "target": x[1] +
                    len(tier.input_string) + index_offset} for x in tier.edges if x[1] != None]
        index_offset += len(tier.input_string)
        symbol_size = min(300 / max(1, len(tier.output_string)), 40)
        colour = shade_colour(colour, (1/steps)*350, g=50, b=20)
        output_x = x + (ind * diff)
        output_y = 300
        outputs = [{'name': f"{x}",
                    'id': f"({ind+index_offset}-{i+index_offset})",
                    'x': output_x,
                    'y': output_y + (i*50),
                    'symbolSize': symbol_size,
                    'label': {'color': contrasting_text_color(colour)},
                    'itemStyle': {'color': colour,
                                  'borderColor': contrasting_text_color(colour)}}
                   for i, x in enumerate(tier.output_string)]
        nodes += outputs
    return nodes, edges


def return_empty_mappings(n=DEFAULT_N):
    ''' Return 'n' * empty mappings
    '''
    y = 0
    mappings = []
    while y < n:
        mappings.append({
            "in": '',
            "out": '',
            "context_before": '',
            "context_after": ''
        })
        y += 1
    return mappings


def hot_to_mappings(hot_data):
    ''' Parse data from HandsOnTable to Mapping format
    '''
    return [{"context_before": str(x[2] or ''), "in": str(x[0] or ''), "context_after": str(x[3] or ''),
             "out": str(x[1] or '')} for x in hot_data if x[0] or x[1]]


def return_descendant_nodes(node: str):
    ''' Return possible outputs for a given input
    '''
    return [x for x in descendants(LANGS_NETWORK, node)]


@APP.route('/')
def home():
    """ Return homepage of g2p studio
    """
    return render_template('index.html')


@APP.route('/docs')
def docs():
    """ Return swagger docs of g2p studio API
    """
    return render_template('docs.html')


@SOCKETIO.on('conversion event', namespace='/convert')
def convert(message):
    """ Convert input text and return output
    """
    transducers = []
    for mapping in message['data']['mappings']:
        mappings_obj = Mapping(hot_to_mappings(mapping['mapping']), abbreviations=flatten_abbreviations(
            mapping['abbreviations']), **mapping['kwargs'])
        transducer = Transducer(mappings_obj)
        transducers.append(transducer)
    transducer = CompositeTransducer(transducers)
    if message['data']['index']:
        tg = transducer(
            message['data']['input_string'])
        data, links = return_echart_data(tg)
        emit('conversion response', {
             'output_string': tg.output_string, 'index_data': data, 'index_links': links})
    else:
        output_string = transducer(
            message['data']['input_string']).output_string
        emit('conversion response', {'output_string': output_string})


@SOCKETIO.on('table event', namespace='/table')
def change_table(message):
    """ Change the lookup table
    """
    if message['in_lang'] == 'custom' or message['out_lang'] == 'custom':
        mappings = Mapping(return_empty_mappings())
    else:
        transducer = make_g2p(message['in_lang'], message['out_lang'])
    if isinstance(transducer, Transducer):
        mappings = [transducer.mapping]
    elif isinstance(transducer, CompositeTransducer):
        mappings = [x.mapping for x in transducer._transducers]
    else:
        pass
    emit('table response', [{'mappings': x.plain_mapping(),
                             'abbs': expand_abbreviations(x.abbreviations),
                             'kwargs': x.kwargs} for x in mappings])


@SOCKETIO.on('connect', namespace='/connect')
def test_connect():
    """ Let client know disconnected
    """
    emit('connection response', {'data': 'Connected'})


@SOCKETIO.on('disconnect', namespace='/connect')
def test_disconnect():
    """ Let client know disconnected
    """
    emit('connection response', {'data': 'Disconnected'})



# -*- coding: utf-8 -*-
"""
This module contains the Transducer and CompositeTransducer classes
which are responsible for performing transductions in the g2p library.
"""

import re
import copy
from typing import Dict, List, Pattern, Tuple, Union
from collections import defaultdict, OrderedDict
from collections.abc import Iterable
from g2p.mappings import Mapping
from g2p.mappings.utils import create_fixed_width_lookbehind, normalize
from g2p.exceptions import MalformedMapping
from g2p.log import LOGGER

# Avoid TypeError in Python < 3.7 (see
# https://stackoverflow.com/questions/6279305/typeerror-cannot-deepcopy-this-pattern-object)
copy._deepcopy_dispatch[type(re.compile(''))] = lambda r, _: r

# An Index is typed as follows:
# {input_index: int, {'input_string': str, 'output': {output_index: int, str}}}
# Example:
# {0: {'input_string': 'h', 'output': {0: 'ʔ'}}}
Index = Dict

# A ChangeLog is a list of changes (List[int])
# The first item (int) in a change is the index of where the change occurs, and the second item (int) is the change offset
# Example:
# an insertion of length 1 at index 0 followed by a deletion of length one at index 2
# [[0,1],[2,-1]]
ChangeLog = List[List[int]]


class TransductionGraph():
    """This is the object returned after performing a transduction using a Transducer.

    Each TransductionGraph must be initialized with an input string.
    """

    def __init__(self, input_string: str):
        # Plain strings
        self._input_string = input_string
        self._output_string = input_string
        # Nodes
        self._input_nodes = [[i, x] for i, x in enumerate(input_string)]
        self._output_nodes = [[i, x] for i, x in enumerate(input_string)]
        # Edges
        self._edges = [[i, i] for i, x in enumerate(input_string)]
        # Debugger
        self._debugger = []

    def __str__(self):
        return self._output_string

    @property
    def input_string(self):
        """str: The input string that initialized the TransductionGraph."""
        return self._input_string

    @input_string.setter
    def input_string(self, value):
        raise ValueError(
            f'Sorry, you tried to change the input string to {value} but it cannot be changed')

    @property
    def output_string(self):
        """str: The output string."""
        return self._output_string

    @output_string.setter
    def output_string(self, value):
        self._output_string = value
        self._output_nodes = [[i, x] for i, x in enumerate(value)]

    @property
    def input_nodes(self):
        """List[List[int, str]]: A list of nodes (index and character string) corresponding to the input"""
        return self._input_nodes

    @input_nodes.setter
    def input_nodes(self, value):
        raise ValueError(
            f'Sorry, you tried to change the input nodes to {value} but they cannot be changed.')

    @property
    def output_nodes(self):
        """List[List[int, str]]: A list of nodes (index and character string) corresponding to the output"""
        return self._output_nodes

    @output_nodes.setter
    def output_nodes(self, value):
        raise ValueError(
            f'Sorry, you tried to change the output nodes to {value} but they cannot be changed directly. Change output_string instead.')

    @property
    def edges(self):
        """List[List[int, int]]: A list of edges (input node index, output node index) corresponding to the indices of the transformation"""
        return self._edges

    @edges.setter
    def edges(self, value):
        self._edges = value

    @property
    def debugger(self):
        """List[dict]: A list of rules applied during the transformation. Useful for debugging."""
        return self._debugger

    @debugger.setter
    def debugger(self, value):
        self._debugger = value

    @property
    def tiers(self):
        """List[TransductionGraph]: A list of TransductionGraph objects for each tier in the graph"""
        return self

    @tiers.setter
    def tiers(self, value):
        raise ValueError(
            f'Sorry, you tried to change the tiers to {value} but they cannot be changed')

    def pretty_edges(self):
        edges = copy.deepcopy(self._edges)
        edges.sort(key=lambda x: x[0])
        for i, edge in enumerate(edges):
            if edge[1] != None:
                edges[i] = [self._input_nodes[edge[0]]
                            [1], self._output_nodes[edge[1]][1]]
            else:
                edges[i] = [self._input_nodes[edge[0]][1], None]
        return edges


class Transducer():
    """This is the fundamental class for performing conversions in the g2p library.

    Each Transducer must be initialized with a Mapping object. The Transducer object can then be called to apply the rules from Mapping on a given input.

    Attributes:
        mapping (Mapping): Formatted input/output pairs using the g2p.mappings.Mapping class.
    """

    def __init__(self, mapping: Mapping):
        self.mapping = mapping
        self.case_sensitive = mapping.kwargs['case_sensitive']
        self.norm_form = mapping.kwargs.get('norm_form', 'none')
        self.out_delimiter = mapping.kwargs.get('out_delimiter', '')
        self._index_match_pattern = re.compile(r'(?<={)\d+(?=})')
        self._char_match_pattern = re.compile(r'[^0-9\{\}]+(?={\d+})', re.U)

    def __repr__(self):
        return f"{self.__class__} between {self.mapping.kwargs.get('in_lang', 'und')} and {self.mapping.kwargs.get('out_lang', 'und')}"

    def __call__(self, to_convert: str, index: bool = False, debugger: bool = False):
        """The basic method to transduce an input. A proxy for self.apply_rules.

        Args:
            to_convert (str): The string to convert.

        Returns:
            TransductionGraph: Returns an object with all the nodes representing input and output characters and their corresponding edges representing the indices of the transformation.
        """
        return self.apply_rules(to_convert)

    @staticmethod
    def _pua_to_index(string: str) -> int:
        """Given a string using with characters in the Supllementary Private Use Area A Unicode block
           Produce the number corresponding to the offset from the beginning of the block.

        Args:
            string (str): The string to convert

        Returns:
            int: The offset from the beginning of the block.
        """
        if string:
            intermediate_ord = ord(string[0])
            return intermediate_ord - 983040
        else:
            return - 1

    def resolve_intermediate_chars(self, output_string):
        ''' Go through all chars and resolve any intermediate characters from the Private Supplementary Use Area
            to their mapped equivalents.
        '''
        indices_seen = defaultdict(int)
        for i, char in enumerate(output_string):
            intermediate_index = self._pua_to_index(char)
            # if not Private Supplementary Use character
            if intermediate_index < 0:
                continue
            else:
                output_char_index = indices_seen[intermediate_index]
                try:
                    output_string = output_string[:i] + \
                        self.mapping[intermediate_index]['out'][output_char_index] + \
                        output_string[i+1:]
                except IndexError:
                    indices_seen[intermediate_index] = 0
                    output_char_index = 0
                    output_string = output_string[:i] + \
                        self.mapping[intermediate_index]['out'][output_char_index] + \
                        output_string[i+1:]
                indices_seen[intermediate_index] += 1
        return output_string

    def update_explicit_indices(self, tg, match, io, intermediate_diff, out_string):
        """ Takes an arbitrary number of input & output strings and their corresponding index offsets.
            It then zips them up according to the provided indexing notation.

            Example:
                A rule that turns a sequence of k\u0313 to 'k might would have a default indexing of k -> ' and \u0313 -> k
                It might be desired though to show that k -> k and \u0313 -> ' and their indices were transposed.
                For this, the Mapping could be given the following: [{'in': 'k{1}\u0313{2}', 'out': "'{2}k{1}"}]
                Indices are found with r'(?<={)\d+(?=})' and characters are found with r'[^0-9\{\}]+(?={\d+})'
        """
        input_char_matches = [x.group()
                              for x in self._char_match_pattern.finditer(io['in'])]
        input_match_indices = [
            x.group() for x in self._index_match_pattern.finditer(io['in'])]
        inputs = {}
        index = 0
        start = match.start() + intermediate_diff
        for i, m in enumerate(input_match_indices):
            for j, char in enumerate(input_char_matches[i]):
                if m in inputs:
                    inputs[m].append({'index': index + start, 'string': char})
                else:
                    inputs[m] = [{'index': index + start, 'string': char}]
                index += 1
        output_char_matches = [
            x.group() for x in self._char_match_pattern.finditer(out_string)]
        output_match_indices = [
            x.group() for x in self._index_match_pattern.finditer(out_string)]
        outputs = {}
        index = 0
        for i, m in enumerate(output_match_indices):
            for j, char in enumerate(output_char_matches[i]):
                if m in outputs:
                    outputs[m].append({'index': index + start, 'string': char})
                else:
                    outputs[m] = [{'index': index + start, 'string': char}]
                index += 1
        out_string = re.sub(re.compile(r'{\d+}'), '', out_string)
        deleted = 0
        for match_index, input_matches in inputs.items():
            try:
                output_matches = outputs[match_index]
            except KeyError:
                output_matches = []
            if len(input_matches) == len(output_matches):
                shortest = input_matches
                longest = output_matches
                process = 'basic'
            elif len(input_matches) < len(output_matches):
                shortest = input_matches
                longest = output_matches
                process = 'insert'
            else:
                shortest = output_matches
                longest = input_matches
                process = 'delete'
            for i, char in enumerate(longest):
                if process == 'basic' or i <= len(shortest) - 1:
                    input_index = input_matches[i]['index'] - intermediate_diff
                    output_index = output_matches[i]['index']
                    # don't allow insertion in basic process
                    if output_index >= len(match.group()) + start:
                        output_index = len(match.group()) + start - 1
                    tg.output_string = tg.output_string[:output_index] + \
                        char['string'] + tg.output_string[output_index+1:]
                    tg.edges = [x for x in tg.edges if x[1] != output_index]
                    tg.edges.append([input_index, output_index])
                else:
                    if process == 'insert':
                        input_index = input_matches[-1]['index'] - \
                            intermediate_diff
                        output_index = output_matches[i]['index']
                        tg.output_string = tg.output_string[:output_index] + \
                            char['string'] + tg.output_string[output_index:]
                        for i, edge in enumerate(tg.edges):
                            if edge[1] != None and edge[1] >= output_index:
                                tg.edges[i][1] += 1
                        tg.edges.append([input_index, output_index])
                    else:
                        input_index = input_matches[i]['index'] - \
                            intermediate_diff
                        if output_matches:
                            output_index = output_matches[-1]['index'] - deleted
                        else:
                            output_index = input_index + intermediate_diff - deleted
                        tg.output_string = tg.output_string[:output_index] + \
                            tg.output_string[output_index + 1:]
                        deleted += 1
                        if len(output_matches) > 0:
                            if [input_index, output_index] not in tg.edges:
                                tg.edges.append([input_index, output_index])
                            tg.edges = [
                                x for x in tg.edges if x[1] != output_index]
                        else:
                            for i, edge in enumerate(tg.edges):
                                if edge[1] != None and edge[1] == output_index:
                                    tg.edges[i][1] = None
                        for i, edge in enumerate(tg.edges):
                            if edge[1] != None and edge[1] > output_index:
                                tg.edges[i][1] -= 1

    def update_default_indices(self, tg, match, intermediate_diff, out_string):
        start = match.start() + intermediate_diff
        in_string = match.group()
        in_length = len(in_string)
        out_length = len(out_string)
        if in_length == out_length:
            for i, char in enumerate(out_string):
                tg.output_string = tg.output_string[:i +
                                                    start] + char + tg.output_string[i+start+1:]
            return
        # default insertion(s)
        elif in_length < out_length:
            longest = out_string
            shortest = in_string
            process = 'insert'
        # default deletion(s)
        else:
            longest = in_string
            shortest = out_string
            process = 'delete'
        # iterate the longest string
        deleted = 0
        last_input_node = start
        last_output_node = start
        for i, char in enumerate(longest):
            # if the shorter string still has that output, keep that index
            if i <= len(shortest) - 1:
                tg.output_string = tg.output_string[:i+start] + \
                    out_string[i] + tg.output_string[i+start+1:]
                last_input_node = i + start
                last_output_node = i + start
            # otherwise...
            else:
                # add a new node and increment each following node
                # log the change in order to update the edges.
                if process == 'insert':
                    # Nodes
                    index_to_add = i + start
                    tg.output_string = tg.output_string[:index_to_add] + \
                        char + tg.output_string[index_to_add:]
                    # Edges
                    # Remove previously deleted and increment
                    for i, edge in enumerate(tg.edges):
                        if edge[1] != None and edge[1] >= index_to_add:
                            tg.edges[i][1] += 1
                    # add edge to index of last input character
                    last_input_node = max(
                        [x[0] for x in tg.edges if x[1] == last_output_node])
                    last_output_node = index_to_add
                    tg.edges.append([last_input_node, index_to_add])
                # delete the node and decrement each following node
                # log the change in order to update the edges.
                else:
                    # Nodes
                    index_to_delete = i + start - deleted
                    tg.output_string = tg.output_string[:index_to_delete] + \
                        tg.output_string[index_to_delete + 1:]
                    deleted += 1
                    # Edges
                    # delete
                    last_input_node = max(
                        [x[0] for x in tg.edges if x[1] == index_to_delete])
                    # if rule is not just a simple deletion,
                    # add an edge between the node and the last output node
                    if out_length > 0:
                        if [last_input_node, last_output_node] not in tg.edges:
                            tg.edges.append(
                                [last_input_node, last_output_node])
                        tg.edges = [x for x in tg.edges if x[1]
                                    != index_to_delete]
                    else:
                        for i, edge in enumerate(tg.edges):
                            if edge[1] != None and edge[1] == index_to_delete:
                                tg.edges[i][1] = None
                    # decrement
                    for i, edge in enumerate(tg.edges):
                        if edge[1] != None and edge[1] > index_to_delete:
                            tg.edges[i][1] -= 1

    def apply_rules(self, to_convert: str):
        # perform any normalization
        if not self.case_sensitive:
            to_convert = to_convert.lower()
        if self.norm_form:
            to_convert = normalize(to_convert, self.norm_form)
        tg = TransductionGraph(to_convert)
        # initialize values
        intermediate_forms = False
        # iterate rules
        for io in self.mapping:
            # Do not allow empty rules
            if not io['in'] and not io['out']:
                continue
            io = copy.deepcopy(io)
            intermediate_diff = 0
            for match in io['match_pattern'].finditer(tg.output_string):
                debug_string = tg.output_string
                start = match.start() + intermediate_diff
                end = match.end() + intermediate_diff
                if 'intermediate_form' in io:
                    out_string = io['intermediate_form']
                    intermediate_forms = True
                else:
                    out_string = io['out']
                if self.out_delimiter:
                    # if not end segment, add delimiter
                    if not end >= len(tg.output_string):
                        out_string += self.out_delimiter
                if any(self._char_match_pattern.finditer(io['in'])) and any(self._char_match_pattern.finditer(out_string)):
                    self.update_explicit_indices(
                        tg, match, io, intermediate_diff, out_string)
                else:
                    self.update_default_indices(
                        tg, match, intermediate_diff, out_string)
                if io['in'] != io['out']:
                    tg.debugger.append({'input': debug_string,
                                        'output': tg.output_string,
                                        'rule': {k: v for k, v in io.items() if k != 'match_pattern'},
                                        'start': start,
                                        'end': end})
                out_string = re.sub(re.compile(r'{\d+}'), '', out_string)
                intermediate_diff += len(out_string) - len(match.group())
        if intermediate_forms:
            tg.output_string = self.resolve_intermediate_chars(
                tg.output_string)
        tg.edges = list(dict.fromkeys(
            [tuple(x) for x in sorted(tg.edges, key=lambda x: x[0])]))
        return tg


class CompositeTransductionGraph(TransductionGraph):
    """This is the object returned after performing a transduction using a CompositeTransducer.

    Each CompositeTransductionGraph must be initialized with a list of TransductionGraph objects.
    """

    def __init__(self, tg_list):
        # Plain strings
        self._input_string = tg_list[0].input_string
        self._output_string = tg_list[-1].output_string
        # Nodes
        self._input_nodes = tg_list[0].input_nodes
        self._output_nodes = tg_list[-1].output_nodes
        # Edges
        self._edges = [x.edges for x in tg_list]
        # Debugger
        self._debugger = [x.debugger for x in tg_list]
        # Tiers
        self._tiers = tg_list

    @property
    def tiers(self):
        """List[TransductionGraph]: A list of TransductionGraph objects for each tier in the CompositeTransducer"""
        return self._tiers

    @tiers.setter
    def tiers(self, value):
        raise ValueError(
            f'Sorry, you tried to change the tiers to {value} but they cannot be changed')

    def pretty_edges(self):
        pretty_edges = []
        for tier_i, edges in enumerate(self._edges):
            edges = copy.deepcopy(edges)
            edges.sort(key=lambda x: x[0])
            for i, edge in enumerate(edges):
                if edge[1] != None:
                    edges[i] = [self.tiers[tier_i].input_nodes[edge[0]]
                                [1], self.tiers[tier_i].output_nodes[edge[1]][1]]
                else:
                    edges[i] = [self.tiers[tier_i].input_nodes[edge[0]][1], None]
            pretty_edges.append(edges)
        return pretty_edges


class CompositeTransducer():
    """This class combines Transducer objects to form a CompositeTransducer object.

    Attributes:
        transducers (List[Transducer]): A list of Transducer objects to compose.
    """

    def __init__(self, transducers: List[Transducer]):
        self._transducers = transducers
        self._tiers = []

    def __repr__(self):
        return f"{self.__class__} between {self._transducers[0].mapping.kwargs.get('in_lang', 'und')} and {self._transducers[-1].mapping.kwargs.get('out_lang', 'und')}"

    def __call__(self, to_convert: str):
        return self.apply_rules(to_convert)

    def apply_rules(self, to_convert: str):
        for transducer in self._transducers:
            tg = transducer(to_convert)
            self._tiers.append(tg)
            to_convert = tg.output_string
        return CompositeTransductionGraph(self._tiers)



"""
    Error Log

"""

import logging
import coloredlogs
import sys

FIELD_STYLES = dict(
    levelname=dict(color='green'),
)

def setup_logger(name):
    """ Create logger and configure with cool colors!
    """

    logging.basicConfig(
        level=logging.INFO
        # filename="logger.log"
    )
    logger = logging.getLogger(name)
    coloredlogs.install(level='INFO', fmt='%(levelname)s - %(message)s',
                        logger=logger, field_styles=FIELD_STYLES)
    return logger

LOGGER = setup_logger('root')



import os
import pickle
from copy import deepcopy
from pathlib import Path
import timeit
import yaml

from networkx import DiGraph, read_gpickle, shortest_path, write_gpickle

from g2p.mappings.utils import find_mapping, load_mapping_from_path
from g2p.exceptions import MappingMissing

LANGS_DIR = os.path.dirname(__file__)
LANGS_PKL = os.path.join(LANGS_DIR, 'langs.pkl')
LANGS_NWORK_PATH = os.path.join(LANGS_DIR, 'network.pkl')

def cache_langs():
    ''' Read in all files and save as pickle
    '''
    langs = {}
    dir_path = Path(LANGS_DIR)
    # Sort by language code
    paths = sorted(dir_path.glob('./*/config.y*ml'), key=lambda x: x.parent.stem)
    mappings_legal_pairs = []
    for path in paths:
        code = path.parent.stem
        with open(path, encoding='utf8') as f:
            data = yaml.safe_load(f)
        # If there is a mappings key, there is more than one mapping
        # TODO: should put in some measure to prioritize non-generated mappings and warn when they override
        if 'mappings' in data:
            for index, mapping in enumerate(data['mappings']):
                mappings_legal_pairs.append((data['mappings'][index]['in_lang'], data['mappings'][index]['out_lang']))
                data['mappings'][index] = load_mapping_from_path(path, index)
        else:
            data = load_mapping_from_path(path)
        langs = {**langs, **{code: data}}
    
    # Save as a Directional Graph
    lang_network = DiGraph()
    lang_network.add_edges_from(mappings_legal_pairs)

    with open(LANGS_NWORK_PATH, 'wb') as f:
        write_gpickle(lang_network, f, protocol=4)

    with open(LANGS_PKL, 'wb') as f:
        pickle.dump(langs, f, protocol=4)

    return langs

# Cache mappings as pickle file for quick loading
with open(LANGS_PKL, 'rb') as f:
    LANGS = pickle.load(f)

LANGS_NETWORK = read_gpickle(LANGS_NWORK_PATH)
LANGS_AVAILABLE = [{k: v['language_name']} for k, v in LANGS.items() if k not in ['generated', 'font-encodings']]
MAPPINGS_AVAILABLE = [mapping for k, v in LANGS.items() for mapping in v['mappings']]

if __name__ == "__main__":
    cache_langs()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

######################################################################
#
# create_ipa_mapping.py
#
# Given two IPA inventories in JSON (either as dedicated inventory
# files or the input/output sides of mapping files), map the first
# onto the second by use of panphon's phonetic distance calculators.
#
# The resulting mappings are used just like other mappings: to make
# converters and pipelines of converters in convert_orthography.py
#
######################################################################

from __future__ import print_function, unicode_literals
from __future__ import division, absolute_import

from typing import List
import argparse
import json
import io

from panphon.xsampa import XSampa
import panphon.distance
from tqdm import tqdm

from g2p.mappings.utils import is_ipa, is_xsampa
from g2p.transducer import Transducer
from g2p.mappings import Mapping
from g2p.log import LOGGER

#################################
#
# Preprocessing:
#
# Panphon can only match a single segment to another segment, rather
# than (say) try to combine two segments to better match the features.
# For example, you might want "kʷ" to match to "kw", but Panphon will
# only match the "kʷ" to "k" and consider the "w" to be a dropped
# character.  In order to get around this, we preprocess strings so
# that common IPA segments that you might expect map to two characters
# in another language, like affricates or rounded consonants, are
# treated as two characters rather than one.
#
#################################


def panphon_preprocess(inventory: List[str], is_xsampa: bool = False):
    xsampa_converter = XSampa()
    panphon_preprocessor = Transducer(Mapping(id='panphon_preprocessor'))
    new_inventory = []
    for x in inventory:
        if is_xsampa:
            x = xsampa_converter.convert(x)
        x = panphon_preprocessor(x)
        new_inventory.append(x)
    return new_inventory

##################################
#
# Creating the mapping
#
#
#
###################################


def create_mapping(l1_mapping: Mapping, l2_mapping: Mapping) -> Mapping:
    ''' Create a mapping from the output of l1 and input of l2.
        Both must be either ipa or x-sampa.
    '''
    l1 = l1_mapping.kwargs['out_lang']
    l2 = l2_mapping.kwargs['in_lang']
    inv_l1 = l1_mapping.inventory("out")
    inv_l2 = l2_mapping.inventory()
    if not is_ipa(l1) and not is_xsampa(l1):
        LOGGER.warning("Unsupported orthography of inventory 1: %s"
                       " (must be ipa or x-sampa)",
                       l1)
    if not is_ipa(l2) and not is_xsampa(l2):
        LOGGER.warning("Unsupported orthography of inventory 2: %s"
                       " (must be ipa or x-sampa)",
                       l2)
    mapping = align_inventories(inv_l1["inventory"], inv_l2["inventory"],
                                is_xsampa(l1), is_xsampa(l2))

    output_mapping = Mapping(mapping, in_lang=l1, out_lang=l2)
    return output_mapping

def find_good_match(p1, inventory_l2, l2_is_xsampa=False):
    """Find a good sequence in inventory_l2 matching p1."""
    # The proper way to do this would be with some kind of beam search
    # through a determinized/minimized FST, but in the absence of that
    # we can do a kind of heurstic greedy search.  (we don't want any
    # dependencies outside of PyPI otherwise we'd just use OpenFST)
    dst = panphon.distance.Distance()
    p1_pseq = dst.fm.ipa_segs(p1)
    p2_pseqs = [dst.fm.ipa_segs(p)
                for p in panphon_preprocess(inventory_l2, l2_is_xsampa)]
    i = 0
    good_match = []
    while i < len(p1_pseq):
        best_input = ""
        best_output = -1
        best_score = 0xdeadbeef
        for j, p2_pseq in enumerate(p2_pseqs):
            # FIXME: Should also consider the (weighted) possibility
            # of deleting input or inserting any segment (but that
            # can't be done with a greedy search)
            if len(p2_pseq) == 0:
                LOGGER.warning('No panphon mapping for %s - skipping',
                               inventory_l2[j])
                continue
            e = min(i + len(p2_pseq), len(p1_pseq))
            input_seg = p1_pseq[i:e]
            score = dst.weighted_feature_edit_distance(''.join(input_seg),
                                                       ''.join(p2_pseq))
            # Be very greedy and take the longest match
            if (score < best_score
                or score == best_score
                    and len(input_seg) > len(best_input)):
                best_input = input_seg
                best_output = j
                best_score = score
        LOGGER.debug('Best match at position %d: %s => %s',
                     i, best_input, inventory_l2[best_output])
        good_match.append(inventory_l2[best_output])
        i += len(best_input)  # greedy!
    return ''.join(good_match)


def align_inventories(inventory_l1, inventory_l2,
                      l1_is_xsampa=False, l2_is_xsampa=False):
    mapping = []
    inventory_l1 = sorted(set(inventory_l1))
    inventory_l2 = list(set(inventory_l2))
    for i1, p1 in enumerate(tqdm(panphon_preprocess(inventory_l1,
                                                    l1_is_xsampa))):
        # we enumerate the strings because we want to save the original string
        # (e.g., 'kʷ') to the mapping, not the processed one (e.g. 'kw')
        good_match = find_good_match(p1, inventory_l2, l2_is_xsampa)
        mapping.append({"in": inventory_l1[i1], "out": good_match})
    return mapping



import os

from unidecode import unidecode

from g2p import make_g2p
from g2p.log import LOGGER
from g2p.mappings import Mapping
from g2p.mappings.create_ipa_mapping import align_inventories
from g2p.mappings.utils import is_ipa, unicode_escape

def align_to_dummy_fallback(mapping: Mapping, io: str = 'in', write_to_file: bool = False, out_dir: str = ''):
    dummy_inventory = ["ɑ", "i", "u", "t", "s", "n"]
    display_name = mapping.kwargs.get('language_name', 'No Language display name in Config')
    config = {'in_lang': mapping.kwargs[f'{io}_lang'], 'out_lang': 'dummy'}
    default_char = 't'
    if is_ipa(mapping.kwargs[f'{io}_lang']):
        mapping = align_inventories(mapping.inventory(io), dummy_inventory)
    else:
        und_g2p = make_g2p('und', 'und-ipa')
        mapping = [{"in": unicode_escape(x), "out": und_g2p(unidecode(x).lower()).output_string} for x in mapping.inventory(io)]
        dummy_list = align_inventories([x['out'] for x in mapping], dummy_inventory)
        dummy_dict = {}
        for x in dummy_list:
            if x['in']:
                dummy_dict[x['in']] = x['out']
                
        for x in mapping:
            try:
                x['out'] = dummy_dict[x['out']]
            except KeyError:
                LOGGER.warn(f"We couldn't guess at what {x['in']} means, so it's being replaced with '{default_char}' instead.")
                x['out'] = default_char       

    config['mapping'] = mapping
    mapping = Mapping(**config)
    if write_to_file:
        if out_dir:
            if os.path.isdir(out_dir):
                mapping.config_to_file(out_dir)
                mapping.mapping_to_file(out_dir)
            else:
                LOGGER.warning(f'{out_dir} is not a directory. Writing to default instead.')
        else:
            mapping.config_to_file()
            mapping.mapping_to_file()
    return mapping

if __name__ == "__main__":
    test = Mapping(in_lang='git', out_lang='git-ipa')
    dummy_config, dummy_mapping = align_to_dummy_fallback(test, write_to_file=True)


'''
All custom Exceptions
'''

# traceback from exceptions that inherit from this class are suppressed
class CommandLineError(Exception):
    """The traceback of all CommandLineError's is supressed when the
    errors occur on the command line to provide a useful command line
    interface.
    """
    def render(self, msg):
        return msg % vars(self)

class MappingMissing(CommandLineError):
    def __init__(self, in_lang, out_lang):
        self.in_lang = in_lang
        self.out_lang = out_lang
    
    def __str__(self):
        return self.render((
            '\n'
            'There is no mapping between the languages "%(in_lang)s" and "%(out_lang)s", please\n'
            'make sure you spelled the name correctly or go to\n'
            'https://g2p-studio.herokuapp.com/api/v1/langs for a list of mappings'
        ))

class InvalidNormalization(CommandLineError):
    def __init__(self, norm):
        self.norm = norm
    
    def __str__(self):
        return self.render((
            '\n'
            'You provided an invalid argument "%(norm)s" to normalize with. \n'
            'Please use "none" or "NFC", "NFKC", "NFD", or "NFKD"\n'
        ))

class MalformedMapping(CommandLineError):
    def __init__(self):
        pass
    
    def __str__(self):
        return self.render((
            '\n'
            'There is something wrong with your mapping. \n'
            'You might be missing some keywords or \n'
            'Not all of the input and output pairs in your mapping have values for "in" and "out"\n'
            'Please refer to the documentation and fix your mapping.'
        ))

class MalformedLookup(CommandLineError):
    def __init__(self):
        pass
    
    def __str__(self):
        return self.render((
            '\n'
            'In order to use a default lookup table, you need to initialize \n'
            'the Mapping class with a dictionary containing two keys: "lang" and "table". \n'
            'Please try again.'
        ))

class IncorrectFileType(CommandLineError):
    def __init__(self, msg):
        self.msg = msg
    def __str__(self):
        return self.render(self.msg)


from unittest import main, TestCase
import os
import json
from g2p.mappings import Mapping
from g2p.transducer import Transducer
from g2p.tests.public import __file__ as public_data
import unicodedata as ud

class MappingTest(TestCase):
    ''' Basic Mapping Test
    '''

    def setUp(self):
        self.test_mapping_no_norm = Mapping([{'in': '\u00e1', 'out': '\u00e1'}, {'in': '\u0061\u0301', 'out': '\u0061\u0301'}], norm_form='none')
        self.test_mapping_norm = Mapping([{'in': '\u00e1', 'out': '\u00e1'}])
        with open(os.path.join(os.path.dirname(public_data), 'git_to_ipa.json'), encoding='utf8') as f:
            self.json_map = json.load(f)
    
    def test_normalization(self):
        self.assertEqual(ud.normalize('NFD', '\u00e1'), self.test_mapping_norm.mapping[0]['in'])
        self.assertNotEqual(self.test_mapping_norm.mapping[0]['in'], '\u00e1')
        self.assertEqual(self.test_mapping_norm.mapping[0]['in'], '\u0061\u0301')
        self.assertEqual(self.test_mapping_no_norm.mapping[0]['in'], '\u00e1')
        self.assertEqual(self.test_mapping_no_norm.mapping[0]['out'], '\u00e1')
        self.assertEqual(self.test_mapping_no_norm.mapping[1]['in'], '\u0061\u0301')
        self.assertEqual(self.test_mapping_no_norm.mapping[1]['out'], '\u0061\u0301')

    def test_json_map(self):
        json_map = Mapping(self.json_map['map'], **{k:v for k,v in self.json_map.items() if k != 'map'})
        self.assertEqual(len(json_map), 34)
        self.assertTrue(json_map.kwargs['in_metadata']['case_insensitive'])

    def test_as_is(self):
        mapping = Mapping([{'in': 'a', "out": 'b'}, {'in': 'aa', 'out': 'c'}], as_is=False)
        mapping_as_is = Mapping([{'in': 'a', "out": 'b'}, {'in': 'aa', 'out': 'c'}])
        transducer = Transducer(mapping)
        transducer_as_is = Transducer(mapping_as_is)
        self.assertEqual(transducer('aa').output_string, 'c')
        self.assertEqual(transducer_as_is('aa').output_string, 'bb')

    def test_case_sensitive(self):
        mapping = Mapping([{'in': 'A', "out": 'b'}], case_sensitive=False)
        mapping_case_sensitive = Mapping([{'in': 'A', "out": 'b'}])
        transducer = Transducer(mapping)
        transducer_case_sensitive = Transducer(mapping_case_sensitive)
        self.assertEqual(transducer('a').output_string, 'b')
        self.assertEqual(transducer_case_sensitive('a').output_string, 'a')
        self.assertEqual(transducer('A').output_string, 'b')

    def test_escape_special(self):
        mapping = Mapping([{'in': '\d', "out": 'digit'}])
        mapping_escaped = Mapping([{'in': '\d', "out": 'b'}], escape_special=True)
        transducer = Transducer(mapping)
        transducer_escaped = Transducer(mapping_escaped)
        self.assertEqual(transducer('1').output_string, 'digit')
        self.assertEqual(transducer('\d').output_string, '\d')
        self.assertEqual(transducer_escaped('1').output_string, '1')
        self.assertEqual(transducer_escaped('\d').output_string, 'b')

    def test_norm_form(self):
        mapping_nfc = Mapping([{'in': 'a\u0301', "out": 'a'}]) # Defaults to NFC
        mapping_nfd = Mapping([{'in': 'a\u0301', "out": 'a'}], norm_form='NFD')
        mapping_none = Mapping([{'in': 'a\u0301', "out": 'a'}], norm_form=False)

        transducer_nfc = Transducer(mapping_nfc)
        transducer_nfd = Transducer(mapping_nfd)
        transducer_none = Transducer(mapping_none)

        self.assertEqual(transducer_nfc('a\u0301').output_string, 'a')
        self.assertEqual(transducer_nfc('\u00E1').output_string, 'a')
        self.assertEqual(transducer_nfd('a\u0301').output_string, 'a')
        self.assertEqual(transducer_nfd('\u00E1').output_string, 'a')
        self.assertEqual(transducer_none('a\u0301').output_string, 'a')
        self.assertEqual(transducer_none('\u00E1').output_string, '\u00E1')

    def test_reverse(self):
        mapping = Mapping([{'in': 'a', "out": 'b'}])
        mapping_reversed = Mapping([{'in': 'a', "out": 'b'}], reverse=True)
        transducer = Transducer(mapping)
        transducer_reversed = Transducer(mapping_reversed)
        self.assertEqual(transducer('a').output_string, 'b')
        self.assertEqual(transducer('b').output_string, 'b')
        self.assertEqual(transducer_reversed('a').output_string, 'a')
        self.assertEqual(transducer_reversed('b').output_string, 'a')

    def test_minimal(self):
        mapping = Mapping(os.path.join(os.path.dirname(public_data), 'mappings', 'minimal_config.yaml'))
        transducer = Transducer(mapping)
        self.assertEqual(transducer('abb').output_string, 'aab')
        self.assertEqual(transducer('a').output_string, 'a')
        self.assertTrue(mapping.kwargs['as_is'])
        self.assertFalse(mapping.kwargs['case_sensitive'])
        self.assertTrue(mapping.kwargs['escape_special'])
        self.assertEqual(mapping.kwargs['norm_form'], 'NFD')
        self.assertTrue(mapping.kwargs['reverse'])

    def test_null_input(self):
        mapping = Mapping([{'in': '', 'out': 'a'}])
        self.assertFalse(mapping())

    def test_no_escape(self):
        mapping = Mapping(os.path.join(os.path.dirname(public_data), 'mappings', 'no_escape.csv'))
        transducer = Transducer(mapping)
        self.assertEqual(transducer('?').output_string, 'ʔ')

if __name__ == "__main__":
    main()


import os
from unittest import main, TestCase
from g2p.mappings import Mapping
from g2p.tests.public import PUBLIC_DIR
from g2p.transducer import CompositeTransducer, Transducer


class TransducerTest(TestCase):
    ''' Basic Transducer Test
    '''
    @classmethod
    def setUpClass(cls):
        cls.test_mapping_moh = Mapping(in_lang="moh", out_lang='moh-ipa')
        cls.test_mapping = Mapping([{'in': 'a', "out": 'b'}])
        cls.test_mapping_rev = Mapping(
            [{"in": 'a', "out": 'b'}], reverse=True)
        cls.test_mapping_ordered_feed = Mapping(
            [{"in": "a", "out": "b"}, {"in": "b", "out": "c"}])
        cls.test_mapping_ordered_counter_feed = Mapping(
            [{"in": "b", "out": "c"}, {"in": "a", "out": "b"}])
        cls.test_as_is_mapping = Mapping(
            [{"in": "j", "out": "ʣ"}, {"in": "'y", "out": "jˀ"}])
        cls.test_not_as_is_mapping = Mapping(
            [{"in": "j", "out": "ʣ"}, {"in": "'y", "out": "jˀ"}], as_is=False)
        cls.test_case_sensitive_mapping = Mapping(
            [{"in": "'n", "out": "n̓"}], case_sensitive=True)
        cls.test_case_insensitive_mapping = Mapping(
            [{"in": "'n", "out": "n̓"}], case_sensitive=False)
        cls.test_case_sensitive_transducer = Transducer(
            cls.test_case_sensitive_mapping)
        cls.test_case_insensitive_transducer = Transducer(
            cls.test_case_insensitive_mapping)
        cls.test_trans_as_is = Transducer(cls.test_as_is_mapping)
        cls.test_trans_not_as_is = Transducer(cls.test_not_as_is_mapping)
        cls.test_trans = Transducer(cls.test_mapping)
        cls.test_trans_ordered_feed = Transducer(
            cls.test_mapping_ordered_feed)
        cls.test_trans_ordered_counter_feed = Transducer(
            cls.test_mapping_ordered_counter_feed)
        cls.test_trans_rev = Transducer(cls.test_mapping_rev)
        cls.test_trans_moh = Transducer(cls.test_mapping_moh)
        cls.test_trans_composite = CompositeTransducer(
            [cls.test_trans, cls.test_trans_rev])
        cls.test_trans_composite_2 = CompositeTransducer(
            [cls.test_trans_rev, cls.test_trans])
        cls.test_regex_set_transducer_sanity = Transducer(
            Mapping([{"in": "a", "out": "b", "context_before": "c"}]))
        cls.test_regex_set_transducer = Transducer(
            Mapping([{"in": "a", "out": "b", "context_before": "[cd]|[fgh]"}]))
        cls.test_deletion_transducer = Transducer(
            Mapping([{'in': 'a', "out": ''}]))
        csv_deletion_mapping = Mapping(os.path.join(
            PUBLIC_DIR, 'mappings', 'deletion_config_csv.yaml'))
        cls.test_deletion_transducer_csv = Transducer(csv_deletion_mapping)
        cls.test_deletion_transducer_json = Transducer(
            Mapping(os.path.join(PUBLIC_DIR, 'mappings', 'deletion_config_json.yaml')))

    def test_ordered(self):
        transducer_i_feed = self.test_trans_ordered_feed('a')
        transducer_feed = self.test_trans_ordered_feed('a')
        transducer_i_counter_feed = self.test_trans_ordered_counter_feed('a')
        transducer_counter_feed = self.test_trans_ordered_counter_feed('a')
        # These should feed b -> c
        self.assertEqual(transducer_feed.output_string, 'c')
        # These should counter-feed b -> c
        self.assertEqual(transducer_counter_feed.output_string, 'b')

    def test_forward(self):
        self.assertEqual(self.test_trans('a').output_string, "b")
        self.assertEqual(self.test_trans('b').output_string, "b")

    def test_backward(self):
        self.assertEqual(self.test_trans_rev("b").output_string, 'a')
        self.assertEqual(self.test_trans_rev("a").output_string, 'a')

    def test_lang_import(self):
        self.assertEqual(self.test_trans_moh('kawennón:nis').output_string, 'kawẽnõːnis')

    def test_composite(self):
        self.assertEqual(self.test_trans_composite('aba').output_string, 'aaa')
        self.assertEqual(self.test_trans_composite_2('aba').output_string, 'bbb')

    def test_as_is(self):
        self.assertEqual(self.test_trans_as_is("'y").output_string, "jˀ")
        self.assertEqual(self.test_trans_not_as_is("'y").output_string, "ʣˀ")

    def test_case_sensitive(self):
        self.assertEqual(self.test_case_sensitive_transducer("'N").output_string, "'N")
        self.assertEqual(self.test_case_sensitive_transducer("'n").output_string, "n̓")
        self.assertEqual(self.test_case_insensitive_transducer("'N").output_string, "n̓")
        self.assertEqual(self.test_case_insensitive_transducer("'n").output_string, "n̓")

    def test_regex_set(self):
        # https://github.com/roedoejet/g2p/issues/15
        self.assertEqual(self.test_regex_set_transducer_sanity('ca').output_string, 'cb')
        self.assertEqual(self.test_regex_set_transducer('ca').output_string, 'cb')
        self.assertEqual(self.test_regex_set_transducer('fa').output_string, 'fb')

    def test_deletion(self):
        self.assertEqual(self.test_deletion_transducer('a').output_string, '')
        self.assertEqual(self.test_deletion_transducer_csv('a').output_string, '')
        self.assertEqual(self.test_deletion_transducer_json('a').output_string, '')


if __name__ == "__main__":
    main()



from unittest import main, TestCase
import os
import csv
from glob import glob

from g2p.app import APP
from g2p.log import LOGGER
from g2p.cli import convert, update
from g2p.tests.public.data import __file__ as data_dir

class CliTest(TestCase):
    def setUp(self):
        self.runner = APP.test_cli_runner()
        DATA_DIR = os.path.dirname(data_dir)
        self.langs_to_test = []
        for fn in glob(f'{DATA_DIR}/*.*sv'):
            if fn.endswith('csv'):
                delimiter = ','
            elif fn.endswith('psv'):
                delimiter = '|'
            elif fn.endswith('tsv'):
                delimiter = '\t'
            with open(fn, encoding="utf-8") as csvfile:
                reader = csv.reader(csvfile, delimiter=delimiter)
                for row in reader:
                    if len(row) != 4:
                        LOGGER.warning(f'Row in {fn} containing values {row} does not have the right values. Please check your data.')
                    else:
                        self.langs_to_test.append(row)

    def test_update(self):
        result = self.runner.invoke(update)
        self.assertEqual(result.exit_code, 0)
    
    def test_convert(self):
        error_count = 0
        for test in self.langs_to_test:
            output_string = self.runner.invoke(convert, [test[2], test[0], test[1]]).stdout.strip()
            if output_string != test[3]:
                LOGGER.warning("test_cli.py: mapping error: {} from {} to {} should be {}, got {}".format(test[2], test[0], test[1], test[3], output_string))
                if error_count == 0:
                    first_failed_test = test
                error_count += 1

        if error_count > 0:
            output_string = self.runner.invoke(convert, [first_failed_test[2], first_failed_test[0], first_failed_test[1]]).stdout.strip()
            self.assertEqual(output_string, first_failed_test[3])

if __name__ == '__main__':
    main()



''' Setup for g2p
'''
from os import path
import datetime as dt
from setuptools import setup, find_packages

build_no = dt.datetime.today().strftime('%Y%m%d')

# Ugly hack to read the current version number without importing g2p:
# (works by )
with open("g2p/_version.py", "r", encoding="utf8") as version_file:
    namespace = {}
    exec(version_file.read(), namespace)
    VERSION = namespace['VERSION'] + "." + build_no

this_directory = path.abspath(path.dirname(__file__))

with open(path.join(this_directory, 'README.md'), encoding='utf8') as f:
    long_description = f.read()

with open(path.join(this_directory, 'requirements.txt'), encoding='utf8') as f:
    REQS = f.read().splitlines()

setup(
    name='g2p',
    python_requires='>=3.6',
    version=VERSION,
    author="Aidan Pine",
    author_email="hello@aidanpine.ca",
    license="MIT",
    url="https://github.com/roedoejet/g2p",
    description='Module for creating context-aware, rule-based G2P mappings that preserve indices',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    include_package_data=True,
    install_requires=REQS,
    entry_points={
        'console_scripts': [
            'g2p = g2p.cli:cli'
    ]},
    zip_safe=False
)



# -*- coding: utf-8 -*-

from unittest import main, TestCase
import os

from networkx.exception import NetworkXNoPath

from g2p.mappings import Mapping
from g2p.transducer import CompositeTransducer, Transducer
from g2p import make_g2p

class NetworkTest(TestCase):
    ''' Basic Test for available networks
    '''

    def setUp(self):
        pass

    def test_not_found(self):
        with self.assertRaises(FileNotFoundError):
            make_g2p('foo', 'eng-ipa')
        with self.assertRaises(FileNotFoundError):
            make_g2p('git', 'bar')

    def test_no_path(self):
        with self.assertRaises(NetworkXNoPath):
            make_g2p('hei', 'git')

    def test_valid_composite(self):
        transducer = make_g2p('atj', 'eng-ipa')
        self.assertTrue(isinstance(transducer, CompositeTransducer))
        self.assertEqual('niɡiɡw', transducer('nikikw').output_string)

    def test_valid_transducer(self):
        transducer = make_g2p('atj', 'atj-ipa')
        self.assertTrue(isinstance(transducer, Transducer))
        self.assertEqual('niɡiɡw', transducer('nikikw').output_string)

if __name__ == "__main__":
    main()



import os
import click
import yaml
from pprint import PrettyPrinter as pp
from flask.cli import FlaskGroup
from collections import OrderedDict
from networkx import draw, has_path

from g2p.transducer import CompositeTransducer, Transducer
from g2p.mappings.create_fallback_mapping import align_to_dummy_fallback
from g2p.mappings.langs import cache_langs, LANGS_NETWORK
from g2p.mappings.create_ipa_mapping import create_mapping
from g2p.mappings.utils import is_ipa, is_xsampa
from g2p.mappings import Mapping
from g2p._version import VERSION
from g2p.app import APP, SOCKETIO, network_to_echart
from g2p.api import update_docs
from g2p.log import LOGGER
from g2p import make_g2p

PRINTER = pp(indent=4)


def create_app():
    return APP


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.version_option(version=VERSION, prog_name="g2p")
@click.group(cls=FlaskGroup, create_app=create_app, context_settings=CONTEXT_SETTINGS)
def cli():
    '''Management script for G2P'''


@click.option('--out-dir', type=click.Path(exists=True, file_okay=False, dir_okay=True))
@click.option('--ipa/--no-ipa', default=False)
@click.option('--dummy/--no-dummy', default=False)
@click.argument('in_lang', type=click.Choice([x for x in LANGS_NETWORK.nodes if not is_ipa(x) and not is_xsampa(x)]))
@cli.command(context_settings=CONTEXT_SETTINGS)
def generate_mapping(in_lang, dummy, ipa, out_dir):
    ''' Generate English mapping.
    '''
    if not ipa and not dummy:
        click.echo('You have to choose to generate either an IPA-based mapping or a dummy fallback mapping. Check the docs for more information.')
    if out_dir and (os.path.exists(os.path.join(out_dir, 'config.yaml')) or os.path.exists(os.path.join(out_dir, 'config.yaml'))):
        click.echo(
            f'There is already a mapping config file in \'{out_dir}\' \nPlease choose another path.')
        return
    if ipa:
        eng_ipa = Mapping(in_lang='eng-ipa', out_lang='eng-arpabet')
        new_mapping = Mapping(in_lang=in_lang, out_lang=f'{in_lang}-ipa')
        click.echo(f"Writing English IPA mapping for {in_lang} to file")
        create_mapping(new_mapping, eng_ipa,
                       write_to_file=True, out_dir=out_dir)
    if dummy:
        new_mapping = Mapping(in_lang=in_lang, out_lang=f'{in_lang}-ipa')
        click.echo(f"Writing dummy fallback mapping for {in_lang} to file")
        dummy_config, dummy_mapping = align_to_dummy_fallback(
            new_mapping, write_to_file=True, out_dir=out_dir)


@click.argument('path', type=click.Path(exists=True, file_okay=False, dir_okay=True))
@cli.command(context_settings=CONTEXT_SETTINGS)
def generate_mapping_network(path):
    ''' Generate a png of the network of mapping languages. Requires matplotlib.
    '''
    import matplotlib.pyplot as plt
    draw(LANGS_NETWORK, with_labels=True)
    plt.show()


@click.option('--debugger/--no-debugger', default=False)
@click.option('--path', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.argument('out_lang')
@click.argument('in_lang')
@click.argument('input_text', type=click.STRING)
@cli.command(context_settings=CONTEXT_SETTINGS)
def convert(in_lang, out_lang, input_text, path, debugger):
    '''Convert from in-lang to out-lang. Visit http://g2p-studio.herokuapp.com/api/v1/langs for a list of options.
    '''
    # Check valid input
    # Check input != output
    if in_lang == out_lang:
        raise click.UsageError(
            "Values must be different for 'IN_LANG' and 'OUT_LANG'")
    # Check input lang exists
    if not in_lang in LANGS_NETWORK.nodes:
        raise click.UsageError(
            f"'{in_lang}' is not a valid value for 'IN_LANG'")
    # Check output lang exists
    if not out_lang in LANGS_NETWORK.nodes:
        raise click.UsageError(
            f"'{out_lang}' is not a valid value for 'OUT_LANG'")
    # Check if path exists
    if not has_path(LANGS_NETWORK, in_lang, out_lang):
        raise click.UsageError(
            f"Path between '{in_lang}' and '{out_lang}' does not exist")
    if os.path.exists(input_text) and input_text.endswith('txt'):
        with open(input_text, encoding='utf8') as f:
            input_text = f.read()
    if in_lang and out_lang:
        transducer = make_g2p(in_lang, out_lang)
    elif path:
        transducer = Transducer(Mapping(path))
    tg = transducer(input_text)
    if debugger:
        output = [tg.output_string, tg.edges, tg.debugger]
        PRINTER.pprint(output)
    else:
        output = tg.output_string
        click.echo(output)


@cli.command(context_settings=CONTEXT_SETTINGS)
def update():
    ''' Update cached language files
    '''
    cache_langs()
    update_docs()
    network_to_echart(write_to_file=True)



""" Organize tests into Test Suites
"""

import os
import sys
from unittest import TestLoader, TextTestRunner, TestSuite

# Unit tests
from g2p.log import LOGGER
from g2p.tests.test_create_mapping import MappingCreationTest
from g2p.tests.test_mappings import MappingTest
from g2p.tests.test_network import NetworkTest
from g2p.tests.test_indices import IndicesTest
from g2p.tests.test_langs import LangTest
from g2p.tests.test_transducer import TransducerTest
from g2p.tests.test_cli import CliTest
from g2p.tests.test_utils import UtilsTest
from g2p.tests.test_fallback import FallbackTest
from g2p.tests.test_api_resources import ResourceIntegrationTest
from g2p.tests.test_studio import StudioTest


LOADER = TestLoader()

TRANSDUCER_TESTS = [
    LOADER.loadTestsFromTestCase(test)
    for test in [
        IndicesTest,
        TransducerTest
    ]
]

MAPPINGS_TESTS = [
    LOADER.loadTestsFromTestCase(test)
    for test in [
        FallbackTest,
        MappingCreationTest,
        MappingTest,
        NetworkTest,
        UtilsTest
    ]
]

LANGS_TESTS = [
    LOADER.loadTestsFromTestCase(test) for test in [
        LangTest,
    ]
]

INTEGRATION_TESTS = [
    LOADER.loadTestsFromTestCase(test) for test in [
        CliTest, ResourceIntegrationTest, StudioTest
    ]
]

DEV_TESTS = TRANSDUCER_TESTS + MAPPINGS_TESTS + LANGS_TESTS + INTEGRATION_TESTS


def run_tests(suite):
    ''' Decide which Test Suite to run
    '''
    if suite == 'all':
        suite = LOADER.discover(os.path.dirname(__file__))
    if suite == 'trans':
        suite = TestSuite(TRANSDUCER_TESTS)
    if suite == 'langs':
        suite = TestSuite(LANGS_TESTS)
    if suite == 'mappings':
        suite = TestSuite(MAPPINGS_TESTS)
    if suite == 'integ':
        suite = TestSuite(INTEGRATION_TESTS)
    elif suite == 'dev':
        suite = TestSuite(DEV_TESTS)
    runner = TextTestRunner(verbosity=3)
    if isinstance(suite, str):
        LOGGER.error("Please specify a test suite to run: i.e. 'dev' or 'all'")
    else:
        return runner.run(suite)
    


if __name__ == "__main__":
    try:
        run_tests(sys.argv[1])
    except IndexError:
        LOGGER.error("Please specify a test suite to run: i.e. 'dev' or 'all'")
